" Copyright 2010 Canonical Ltd.  This software is licensed under the
" GNU Affero General Public License version 3 (see the file LICENSE).
"
" Python function to format a python from/import statement by sorting
" the imported items and wrapping the lines.
" Language:     Python (ft=python)
" Last Change:  Wed May 19 09:06:18 CDT 2010
" Maintainer:   Edwin Grubbs <edwin.grubbs@canonical.com>
" Version:      0.1, for Vim 7.2
" URL:

if exists('b:loaded_python_import_macros')
    finish
endif
let b:loaded_python_import_macros = 1

" Add canonical-vim/share/python_path to the PYTHONPATH.
runtime canonical-vim/share/python_path.vim

highlight link NewImportID Todo

python << EOF
import os
import re
import string
import vim

def LP_inputlist(options):
    """Wrap the inputlist() vim function.

    This allows tests to provide a mock function, since vim doesn't allow
    overriding built-in functions, and runVimTests is not able to automate
    input to the inputlist() prompt.
    """
    # vim.eval() can't deal with a double-quote inside of single
    # quotes, but it can if double-quotes are escaped inside of
    # double-quotes.
    options = ['"%s"' % option.replace('"', r'\"')
               for option in options]
    return int(vim.eval('inputlist([%s])' % ', '.join(options)))

def LP_format_python_import():
    MAX_LINE_LENGTH = 78
    line_number, column_number = vim.current.window.cursor
    line_number -= 1 # Convert from 1-indexed to 0-indexed

    line = vim.current.buffer[line_number]
    match = re.match(r'(\s*)from\s+[A-Za-z_][A-Za-z0-9_.]*\s+import\s+', line)
    if match is None:
        print 'from/import statement not found: %r' % line
        return

    from_indent = match.group(1)

    # Fill the 'statement' variable with all the text for the
    # from/import statement.
    statement = line
    statement_size = 1
    if '(' not in statement:
        # Single line import statement.
        from_expr = r'(.*import *)'
        full_expr = from_expr + r'(.*)'
    else:
        # Multi line import statement.
        for i in range(line_number+1, len(vim.current.buffer)):
            if ')' in statement:
                break
            statement_size += 1
            statement += vim.current.buffer[i]

        from_expr = r'(.*import[^( ]*) *\('
        full_expr = from_expr + r'(.*)\)'

    # Parse the from/import statement.
    match = re.match(full_expr, statement)
    if match is None:
        print "Error processing this import"
        return

    from_part, module_part = match.groups()
    from_part = from_part.rstrip() + ' '

    # Sort and remove duplicates from the imported items.
    modules = set(module.strip() for module in module_part.split(','))
    modules = sorted(modules, key=str.lower)

    # Split the imported module names into lines to fall under the
    # MAX_LINE_LENGTH.
    def sum(x, y):
        return x + y
    module_indent = from_indent + (' ' * 4)
    module_groups = [[modules[0]]]
    last_group = module_groups[-1]
    for module in modules[1:]:
        group_length = reduce(sum, [len(i) for i in last_group])
        comma_length = len(', ') * len(last_group) + len(',')
        line_length = (len(module_indent) + group_length
                    + comma_length + len(module))
        if line_length > 78:
            last_group = [module]
            module_groups.append(last_group)
        else:
            last_group.append(module)

    # Determine whether the from/import statement fits in one line.
    if (len(module_groups) == 1
        and (len(from_part) + len(', '.join(module_groups[0])) <= 78)):
        new_statement_size = 1
    else:
        new_statement_size = len(module_groups) + 1

    # If the newly formatted from/import statement will have more lines
    # than before (size_diff>0), insert lines into the buffer. If it will
    # have less lines, clear the text from the extra lines of the old
    # from/import statement.
    size_diff = new_statement_size - statement_size
    if size_diff > 0:
        after = line_number + statement_size
        vim.current.buffer[after:after] = [''] * size_diff
    elif size_diff < 0:
        for i in range(new_statement_size, statement_size):
            vim.current.buffer[line_number+i] = ''

    # Add the new from/import statement without parens if it's a single
    # line or with parens for multiple lines.
    if new_statement_size == 1:
        vim.current.buffer[line_number] = (
            from_part + ', '.join(module_groups[0]))
        return (line_number, line_number)
    else:
        vim.current.buffer[line_number] = from_part + '('
        for i in range(0, len(module_groups)):
            current_line = line_number + i + 1
            new_line = module_indent + ', '.join(module_groups[i])
            if i+1 == len(module_groups):
                new_line += ')'
            else:
                new_line += ','
            vim.current.buffer[current_line] = new_line
        return (line_number, line_number+len(module_groups))

def LP_python_import_search():
    valid_chars = '_' + string.ascii_letters + string.digits

    line_number, column_number = vim.current.window.cursor
    line_number -= 1 # Convert from 1-indexed to 0-indexed

    line = vim.current.buffer[line_number]
    if line[column_number] not in valid_chars:
        print '"%s" is not part of an identifier' % line[column_number]

    for start in range(column_number, -1, -1):
        if line[start] not in valid_chars:
            start += 1
            break
    for stop in range(column_number, len(line)):
        if line[stop] not in valid_chars:
            break
    else:
        # If it hits the end of the line, the stop index is actually one more.
        stop += 1

    identifier = line[start:stop]
    try:
        result = _find_import_modules(identifier)
    except LookupError:
        result = []
    ordered_result = sorted(item for item in result if item['module'])
    if len(ordered_result) == 0:
        print "%r not found" % identifier
    else:
        options = ['Select import:']
        for i, item in enumerate(ordered_result):
            info = item['match'] or item['kind']
            option = '%s: %-40s %-38s' % (i+1, item['module'], info)
            options.append(option)

        selected = LP_inputlist(options)
        if selected == 0:
            # Selection canceled.
            pass
        else:
            index = selected - 1
            _insert_import_statement(
                ordered_result[index]['module'], identifier)

def _find_init_file(dirname):
    for extension in ('py', 'pyc', 'pyo'):
        init_file = os.path.join(
            dirname, '__init__.%s' % extension)
        if os.path.isfile(init_file):
            return init_file
    return None

class LP_UniqueModuleDict(dict):
    def __hash__(self):
        return hash(self['module'])

def _find_import_modules(tag):
    match_list = vim.eval("taglist('^%s$')" % tag)
    import_list = []
    for match in match_list[:20]:
        # Determine module name.
        filename = match['filename']
        module_parts = []
        if not os.path.exists(filename):
            print "Filename does not exist:", filename
        while True:
            if os.path.isfile(filename):
                module = os.path.basename(os.path.splitext(filename)[0])
            elif os.path.isdir(filename):
                if _find_init_file(filename) is None:
                    break
                module = os.path.basename(filename)
            else:
                # The tags file has references to files that don't
                # exist.
                break
            module_parts.insert(0, module)
            filename = os.path.dirname(filename)
        if tag in match['cmd']:
            # Remove "/^" and "$/" around the pattern.
            match_definition = match['cmd'][2:-2].strip()
        else:
            if os.path.isfile(filename):
                lines = open(filename).readlines()
            elif os.path.isdir(filename):
                init_file = _find_init_file(filename)
                if init_file is not None:
                    lines = open(filename).readlines()
                else:
                    lines = []
            else:
                lines = []

            # Try to find the definition in the file and look five
            # lines up for the start of the definition.
            match_definition = None
            if lines:
                # Convert to regexp to extended regexp.
                pattern = match['cmd'].replace('(', r'\(')
                pattern = pattern.replace(')', r'\)')
                regex = re.compile(pattern)
                for i, line in enumerate(lines):
                    if regex.match(line):
                        break
                for j in range(i, max(0, i-5), -1):
                    if tag in lines[j]:
                        match_definition = lines[j].strip()

        import_info = LP_UniqueModuleDict(
            module='.'.join(module_parts),
            match=match_definition,
            kind=match['kind'],
            )
        import_list.append(import_info)
    return set(import_list)

def _insert_import_statement(module, identifier):
    # If the from-module statement already exists in the file, add the
    # identifier to that instead of creating a separate import
    # statement.
    half_import_statement = 'from %s ' % module
    for i, line in enumerate(vim.current.buffer):
        if line.startswith(half_import_statement):
            if '(' in line:
                vim.current.buffer[i] = line + ' ' + identifier + ','
            else:
                vim.current.buffer[i] = line + ', ' + identifier
            vim.current.window.cursor = (i+1, 0)
            buffer_start, buffer_stop = LP_format_python_import()
            LP_highlighter.highlight_imported_identifier(
                identifier, buffer_start, buffer_stop)
            return

    # There is no existing import statement to add the identier to, so
    # determine where it should be placed in the file.
    # The `place` variable is set back to None whenever statements
    # are found that should be above the import statement.
    top_level_module = half_import_statement.split('.')[0]
    place = None
    for i, line in enumerate(vim.current.buffer):
        if line == '':
            if place is None:
                # The first blank line in the file or the first blank line
                # after __metaclass__, __all__, or an import would be good,
                # but there might be a better match, so don't break yet.
                place = i + 1
        elif ('__metaclass__' in line
              or '__all__' in line
              or 'import ' in line):
            # The next blank line would be a good place.
            place = None
        if line.startswith(top_level_module):
            # Ideal position.
            place = i
            break
    if place is None:
        # It is normally only possible to get here if the file has no
        # empty lines.
        place = 0
    import_statement = half_import_statement + 'import ' + identifier
    vim.current.buffer[place:place] = ['', import_statement, '']
    vim.current.window.cursor = (place+2, 0)

    buffer_start, buffer_stop = LP_format_python_import()
    LP_highlighter.highlight_imported_identifier(
        identifier, buffer_start, buffer_stop)

class LP_HighlighterClass:
    def __init__(self):
        self.matches = []

    def unhighlight(self):
        while len(self.matches) > 0:
            vim.eval("matchdelete('%s')" % self.matches.pop())

    def highlight_imported_identifier(
        self, identifier, buffer_start, buffer_stop):
        """Highlight new identifier in import statement."""
        self.unhighlight()
        # Convert buffer_start/stop zero-indexed values to 1-indexed line
        # numbers.
        lineno_above = buffer_start
        lineno_below = buffer_stop + 2
        not_followed_by_import = r'\(.* import \)\@!'

        pattern = ((r'\%>' + str(lineno_above) + r'l')
                + (r'\%<' + str(lineno_below) + r'l')
                + '\(^\|\A\)' + identifier + '\(\A\|$\)'
                + not_followed_by_import)
        # The string literals passed into matchadd() should have single quotes
        # around the pattern so that it doesn't remove backslashes.
        match_id = vim.eval("matchadd('NewImportID', '%s')" % pattern)
        self.matches.append(match_id)


LP_highlighter = LP_HighlighterClass()
vim.command(
    'autocmd CursorHold,WinLeave,BufLeave,InsertEnter <buffer> '
    'python LP_highlighter.unhighlight()')

EOF
